import torch
import argparse
from rknnlite.api import RKNNLite
import sys
import numpy as np
import onnxruntime as ort
import time
from transformers import AutoTokenizer
from IndicTransToolkit.processor import IndicProcessor


class IndicTrans:
    DEVICE = "cpu"
    
    langToCode = {
        "Assamese": "asm_Beng",
        "Kashmiri (Arabic)": "kas_Arab",
        "Punjabi": "pan_Guru",
        "Bengali": "ben_Beng",
        "Kashmiri (Devanagari)": "kas_Deva",
        "Sanskrit": "san_Deva",
        "Bodo": "brx_Deva",
        "Maithili": "mai_Deva",
        "Santali": "sat_Olck",
        "Dogri": "doi_Deva",
        "Malayalam": "mal_Mlym",
        "Sindhi (Arabic)": "snd_Arab",
        "English": "eng_Latn",
        "Marathi": "mar_Deva",
        "Sindhi (Devanagari)": "snd_Deva",
        "Konkani": "gom_Deva",
        "Manipuri (Bengali)": "mni_Beng",
        "Tamil": "tam_Taml",
        "Gujarati": "guj_Gujr",
        "Manipuri (Meitei)": "mni_Mtei",
        "Telugu": "tel_Telu",
        "Hindi": "hin_Deva",
        "Nepali": "npi_Deva",
        "Urdu": "urd_Arab",
        "Kannada": "kan_Knda",
        "Odia": "ory_Orya"
    }
    
    def __init__(self, preEncoderPath, encoderPath, preDecoderPath, decoderPath, modelName = "ai4bharat/indictrans2-en-indic-dist-200M"):
        self.encoderModel = RKNNLite()
        self.decoderModel = RKNNLite()
        self.preEncoderSession = ort.InferenceSession(preEncoderPath)
        self.preDecoderSession = ort.InferenceSession(preDecoderPath)
        self.modelName = modelName
        ret = self.encoderModel.load_rknn(encoderPath)
        if ret != 0:
            print('Load RKNN Model failed, please verify path and model!')
            exit(ret)
            
        ret = self.decoderModel.load_rknn(decoderPath)
        if ret != 0:
            print('Load RKNN Model failed, please verify path and model!')
            exit(ret)
            
        self.ip = IndicProcessor(inference=True)
        
        self.tokenizer = AutoTokenizer.from_pretrained(self.modelName, trust_remote_code=True)
        
        ret = self.encoderModel.init_runtime(core_mask=RKNNLite.NPU_CORE_AUTO)
        if ret!=0:
            print("init runtime failed")
            exit(ret)

        ret = self.decoderModel.init_runtime(core_mask=RKNNLite.NPU_CORE_AUTO)
        if ret!=0:
            print("init runtime failed")
            exit(ret)


    def translate(self, input, srcLang, tgtLang):
        try:
            srcLang = IndicTrans.langToCode[srcLang]
            tgtLang = IndicTrans.langToCode[tgtLang]
        except:
            return "Error: src or target language not found!"
        
        batch = self.ip.preprocess_batch([input], src_lang=srcLang, tgt_lang=tgtLang)

        inputs = self.tokenizer(batch, return_tensors="pt", padding=False, truncation=True, max_length=24, return_attention_mask=True)
        
        input_ids = inputs["input_ids"]
        
        ids = input_ids.numpy()
        n = int(ids.shape[1])
        
        if(n > 8):
            if(n > 16):
                if(n > 24):
                    return "Error: max token length of 20 exceeded for input"
                else:
                    m = 24-n
            else:
                m = 16-n
        else:
            m = 8-n
                    
        if(m>0):
            ons = np.full((1,m+1), fill_value=5, dtype=np.int64)
            ons[0][-1] = 2
            ids = np.concatenate((ids[:, :n-1], ons), axis=1)
          
        inputs = {"input_ids": ids}
        outputs0 = self.preEncoderSession.run(None, inputs)
        enc_outputs = self.encoderModel.inference(inputs=outputs0)
        context_embeddings = torch.tensor(enc_outputs[0], dtype=torch.float16)
        
        decoderInputIds = torch.tensor([[2]])
        max_new_tokens = 24
        rem = int(decoderInputIds.shape[1])
        generated_tokens = []
        iter = 0
        for _ in range(max_new_tokens):        
            iter += 1
            decoderInput = {self.preDecoderSession.get_inputs()[0].name: decoderInputIds.numpy()}
            outputs = self.preDecoderSession.run(None, decoderInput)
            outputs.append(enc_outputs[0])
            dec_outputs = self.decoderModel.inference(inputs=outputs)
            logits = torch.tensor(dec_outputs[0])
            
            next_token_logits = logits[:, -1, :]  
            next_token_id = torch.argmax(next_token_logits, dim=-1).unsqueeze(-1)  
            
            decoderInputIds = torch.cat((decoderInputIds, next_token_id), dim=1)
            generated_tokens.append(next_token_id)
            
            if next_token_id.item() == self.tokenizer.eos_token_id:
                break
            
        output_ids = torch.tensor(generated_tokens, dtype=torch.int64)
        translated_text = self.tokenizer.decode(output_ids, skip_special_tokens=True)
        decoded_text = self.ip.postprocess_batch([translated_text], lang=tgtLang)
        return decoded_text[0]
    
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-pe",
        default = "/opt/vicharak-chat/models/encoder_enin.onnx"
    )
    parser.add_argument(
        "-e",
        default = "/opt/vicharak-chat/models/encoder_enin.rknn"
    )
    
    parser.add_argument(
        "-pd",
        default = "/opt/vicharak-chat/models/decoder_enin.onnx"
    )
    parser.add_argument(
        "-d",
        default = "/opt/vicharak-chat/models/decoder_enin.rknn"
    )
    
    args = parser.parse_args()
    try:
        modelName = "ai4bharat/indictrans2-en-indic-dist-200M"
        model = IndicTrans(args.pe, args.e, args.pd, args.d, modelName)
    except Exception as e:
        print("Error: ", e)
        exit(0)
    
    print("MODEL_LOADED")
    while(True):
        srcLang = input()
        if(srcLang == "/exit"):
            break
        
        tgtLang = input()
        sentence = input()
        
        translation = model.translate(sentence, srcLang, tgtLang)
        
        print(translation)
        